
//
// Licensed to the Apache Software Foundation (ASF) under one
// or more contributor license agreements.  See the NOTICE file
// distributed with this work for additional information
// regarding copyright ownership.  The ASF licenses this file
// to you under the Apache License, Version 2.0 (the
// "License"); you may not use this file except in compliance
// with the License.  You may obtain a copy of the License at
// 
//   http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.
//

// This source file was created by a code generator.
// Please do not edit.

#include "qpid/log/Statement.h"
#include "qpid/framing/FieldTable.h"
#include "qpid/management/Manageable.h" 
#include "qpid/agent/ManagementAgent.h"
#include "Agent.h"


using namespace qpid::management;
using namespace qpid::sys;
using namespace qpid::framing;
using           std::string;

string  Agent::packageName  = string ("qpid");
string  Agent::className    = string ("agent");
uint8_t Agent::md5Sum[16]   =
    {0xba,0x80,0xd0,0x67,0x55,0x30,0x2b,0x8a,0x57,0x53,0xfe,0x12,0x18,0xac,0x68,0x5f};

Agent::Agent (ManagementAgent* _agent, Manageable* _core) :
    ManagementObject(_agent, _core)
{
    


}

Agent::~Agent ()
{

}

namespace {
    const string NAME("name");
    const string TYPE("type");
    const string ACCESS("access");
    const string INDEX("index");
    const string UNIT("unit");
    const string MIN("min");
    const string MAX("max");
    const string MAXLEN("maxlen");
    const string DESC("desc");
    const string ARGCOUNT("argCount");
    const string ARGS("args");
    const string DIR("dir");
    const string DEFAULT("default");
}

void Agent::writeSchema (Buffer& buf)
{
    FieldTable ft;

    // Schema class header:
    buf.putShortString (packageName); // Package Name
    buf.putShortString (className);   // Class Name
    buf.putBin128      (md5Sum);      // Schema Hash
    buf.putShort       (5); // Config Element Count
    buf.putShort       (0); // Inst Element Count
    buf.putShort       (0); // Method Count
    buf.putShort       (0); // Event Count

    // Properties
    ft = FieldTable ();
    ft.setString (NAME,   "connectionRef");
    ft.setInt    (TYPE,   TYPE_REF);
    ft.setInt    (ACCESS, ACCESS_RO);
    ft.setInt    (INDEX,  1);
    buf.put (ft);

    ft = FieldTable ();
    ft.setString (NAME,   "label");
    ft.setInt    (TYPE,   TYPE_SSTR);
    ft.setInt    (ACCESS, ACCESS_RO);
    ft.setInt    (INDEX,  0);
    ft.setString (DESC,   "Label for agent");
    buf.put (ft);

    ft = FieldTable ();
    ft.setString (NAME,   "registeredTo");
    ft.setInt    (TYPE,   TYPE_REF);
    ft.setInt    (ACCESS, ACCESS_RO);
    ft.setInt    (INDEX,  0);
    ft.setString (DESC,   "Broker agent is registered to");
    buf.put (ft);

    ft = FieldTable ();
    ft.setString (NAME,   "systemId");
    ft.setInt    (TYPE,   TYPE_UUID);
    ft.setInt    (ACCESS, ACCESS_RO);
    ft.setInt    (INDEX,  0);
    ft.setString (DESC,   "Identifier of system where agent resides");
    buf.put (ft);

    ft = FieldTable ();
    ft.setString (NAME,   "objectIdBank");
    ft.setInt    (TYPE,   TYPE_U32);
    ft.setInt    (ACCESS, ACCESS_RO);
    ft.setInt    (INDEX,  0);
    ft.setString (DESC,   "Assigned object-id bank");
    buf.put (ft);


    // Statistics

    // Methods

    // Events

}



void Agent::writeProperties (Buffer& buf)
{
    sys::Mutex::ScopedLock mutex(accessLock);
    configChanged = false;

    writeTimestamps (buf);
    buf.putLongLong(connectionRef);
    buf.putShortString(label);
    buf.putLongLong(registeredTo);
    systemId.encode(buf);
    buf.putLong(objectIdBank);

}

void Agent::writeStatistics (Buffer& buf, bool skipHeaders)
{
    sys::Mutex::ScopedLock mutex(accessLock);
    instChanged = false;



    if (!skipHeaders)
        writeTimestamps (buf);


    // Maintenance of hi-lo statistics


}

void Agent::doMethod (string, Buffer&, Buffer& outBuf)
{
    Manageable::status_t status = Manageable::STATUS_UNKNOWN_METHOD;

    outBuf.putLong        (status);
    outBuf.putShortString (Manageable::StatusText (status));
}

